/*********************************************************/
/***                 Listing Four                      ***/
/***                  "twain.cpp"                      ***/
/***       This is the interface class for TWAIN       ***/
/***                complient devices.                 ***/
/***                                                   ***/
/***                    written by                     ***/
/***                 Craig A. Lindley                  ***/
/***                                                   ***/
/***      Revision: 1.0   Last Update: 12/11/93        ***/
/*********************************************************/

/*
Revision History

1. 12/11/93 CAL - Code baselined.
*/

#include <mem.h>
#include <string.h>
#include "twain.h"
#include "twain.hpp"

// Variables global to this file only
static HINSTANCE hDLLInst;
static char szAppName []  = "TWAINBackDropWindow";
static Twain *lpTwain;

// Constructor for class
Twain::Twain(void) {

  lpTwain = this;                 // Save a ptr to this object
  hDLLInstance = hDLLInst;        // Store this DLLs' instance in class data

  // Clear the identity structure for application and source.
  memset(&AppIdentity,    0, sizeof(DC_IDENTITY));
  memset(&SourceIdentity, 0, sizeof(DC_IDENTITY));

  // Now set the fields of the structure appropriately. Some fields are
  // hardcoded, some are read as string resources.
  AppIdentity.Id = 0;             // Source Manager will assign real value
  AppIdentity.Version.MajorNum = 1;
  AppIdentity.Version.MinorNum = 0;
  AppIdentity.Version.Language = DCLG_USA;
  AppIdentity.Version.Country  = DCCY_USA;
  AppIdentity.ProtocolMajor = DCON_PROTOCOLMAJOR;
  AppIdentity.ProtocolMinor = DCON_PROTOCOLMINOR;
  AppIdentity.SupportedGroups = DG_IMAGE | DG_CONTROL;

  // Read these strings from resource.
  LoadString(hDLLInstance, IDS_MANUFACTURE,   (LPSTR) &AppIdentity.Manufacturer, sizeof(DC_STR32));
  LoadString(hDLLInstance, IDS_PRODUCTFAMILY, (LPSTR) &AppIdentity.ProductFamily, sizeof(DC_STR32));
  LoadString(hDLLInstance, IDS_PRODUCTNAME,   (LPSTR) &AppIdentity.ProductName, sizeof(DC_STR32));
  LoadString(hDLLInstance, IDS_VERSIONINFO,   (LPSTR) &AppIdentity.Version.Info, sizeof(DC_STR32));
                                             
  DSMOpen = FALSE;
  DSOpen  = FALSE;
  DSEnabled = FALSE;
  hAppWindow = 0;
  XferMechanism = DCSX_NATIVE;
}


Twain::~Twain(void) {

  // Nothing yet to do

}


/*************************************************************************
*FUNCTION:TransferImage
*
*ARGS:    hWindow  Handle of window messages go to
*
*RETURNS: none
*
*NOTES:   1) Get information about image.
*         2) Set up a loop to pull image(s) from the Source
*         3) Call GetCompleteImage to get image
*
*NOTES: This function is called in response to a XFERREADY message.
*************************************************************************/

void Twain::TransferImage(HWND hWindow) {
  DC_IMAGEINFO    dcImageInfo;
  DC_PENDINGXFERS dcPendingXfer;
  DC_UINT16       Status;

  // NOTE: all negotiations are done before image is scanned.

  // Get the image information. It is nice to know a little about the
  // image the Source will be sending. Whether this info is used or not
  // is another matter completely.
  Status = (*lpDSM_Entry)(&AppIdentity,
                          &SourceIdentity,
                          DG_IMAGE,
                          DAT_IMAGEINFO,
                          MSG_GET,
                          (DC_MEMREF) &dcImageInfo);

  if (Status == DCRC_SUCCESS) {
    // Get an image from the Source and then ask if more images are coming.
    // EV only supports one image per transfer. The do/while loop is
    // included for future expansion possibilities.

    // explicitly initialize the count of images Source has for App.
    dcPendingXfer.Count = 0;
    do {
      GetCompleteImage (hWindow); // Transfer the image
      // Required for proper 7<->6 or 5 state transitions
      if ((*lpDSM_Entry)(&AppIdentity,
                         &SourceIdentity,
                         DG_CONTROL,
                         DAT_PENDINGXFERS,
                         MSG_ENDXFER,
                         (DC_MEMREF)&dcPendingXfer) != DCRC_SUCCESS)

        dcPendingXfer.Count = 0;  // trash remaining images

    } while (dcPendingXfer.Count != 0);
  }
}

/*************************************************************************
*FUNCTION:GetCompleteImage
*
*ARGS:    hWindow  Handle to window messages go to
*
*RETURNS: TRUE if success.
*
*NOTES:   Asks the Source to start sending the pending image.  We
*         were told about this by receipt of a XFERREADY message
*         from the Source.
*
*NOTES: For a file transfer only one operation is required
*so the WHILE loop is not required here.  For memory
*transfers, slice/tile by slice/tile the loop would be required as
*well as code to support the additional listed DCRC_XXX values.  The
*additional DCRC_XXX values are DCRC_CANCEL, user cancelled an in
*progress image; DCRC_FAILURE, transfer crashed; DCRC_SUCCESS got a
*slice/tile continue.
*************************************************************************/

BOOL Twain::GetCompleteImage (HWND hWindow) {
  DC_UINT16 Status = DCRC_SUCCESS;
  BOOL Result = FALSE;

  // Go into a loop to get all of the image data
  while (Status != DCRC_XFERDONE) {
    // Ask for image transfer
    Status = (*lpDSM_Entry)(&AppIdentity,
                            &SourceIdentity,
                            DG_IMAGE,
                            DAT_IMAGEFILEXFER,
                            MSG_GET,
                            (DC_MEMREF) NULL);
    // See what the returned status is
    switch (Status) {
      case DCRC_XFERDONE:
        // Transfer is done, pass the acknowledgement back to App
        // thru message queue.
        SendMessage (hWindow, WM_XFERDONE, TRUE, 0);
        Result = TRUE;
        break;

      case DCRC_CANCEL:
      case DCRC_FAILURE:
      default:
        // Something wrong, pass a FALSE back to App
        SendMessage (hWindow, WM_XFERDONE, FALSE, 0);
        Result = FALSE;
    }
  }
  return Result;
}

/*************************************************************************
*FUNCTION:ProcessDCMessage
*
*ARGS:	  lpMsg  Pointer to Windows msg retrieved by GetMessage
*         hWnd   Application's main window handle
*
*RETURNS: TRUE  if application should process message as usual or FALSE
*               if application should skip processing of this message
*
*NOTES:
*************************************************************************/

BOOL Twain::ProcessDCMessage(LPMSG lpMsg, HWND hWindow) {
  DC_UINT16 Status = DCRC_NOTDSEVENT;
  DC_EVENT  dcEvent;


  // Only ask Source Manager to process event if there is a Source connected.
  if (DSMOpen && DSOpen) {
    /* A Source provides a modeless dialog box as its user interface.
     * The following call relays Windows messages down to the Source's
     * UI that were intended for its dialog box. It also retrieves TWAIN
     * messages sent from the Source to the App.
     */
    dcEvent.pEvent = (DC_MEMREF) lpMsg;
    Status = (*lpDSM_Entry)(&AppIdentity,
                            &SourceIdentity,
                            DG_CONTROL,
                            DAT_EVENT,
                            MSG_PROCESSEVENT,
                            (DC_MEMREF) &dcEvent);

    switch (dcEvent.DCMessage) {
      case MSG_XFERREADY:         // Source has an image to xfer
        TransferImage(hWindow);
        break;

      case MSG_CLOSEDSREQ:        // User requested close of UI
        DisableDS();              // Close down UI
        CloseDS();                // Close the Source
        CloseDSM();               // Close the DSM
        DestroyWindow(hBackgroundWindow);
        break;

      case MSG_NULL:              // All other messages go here and
      default:                    // into the bit bucket
        break;
    }
  }
  // tell the caller what happened
  return (Status == DCRC_DSEVENT);
}

/*************************************************************************
*FUNCTION:ScanImage
*
*ARGS:    FileName  Path and FileName without extension to store the
*                   scanned image in.
*RETURNS: TRUE if successful, FALSE if not
*
*NOTES:
************************************************************************/

BOOL Twain::ScanImage(LPSTR ImageFileName) {

  lpFileName = ImageFileName;     // Save ptr to filename
  ScanResults = FALSE;            // No image scanned yet

  // Create TWAIN backdrop window. It is parent to Source's UI dialog
  hBackgroundWindow = CreateWindow(szAppName, NULL,
                        WS_POPUP | WS_BORDER,
                        0, 0, CW_USEDEFAULT, CW_USEDEFAULT,
                        HWND_DESKTOP, NULL, hDLLInstance, NULL);

  ShowWindow(hBackgroundWindow, SW_SHOWMAXIMIZED);
  SendMessage(hBackgroundWindow, WM_SYSCOMMAND, SC_MAXIMIZE, 0L); // Expand parent window to full screen

  // The modeless Source UI window must be taken into consideration
  MSG msg;
  while (GetMessage (&msg, NULL, 0, 0))  {
    if ((!DSOpen) || (!ProcessDCMessage((LPMSG)&msg, hBackgroundWindow))) {
      TranslateMessage (&msg);
      DispatchMessage (&msg);
    }
  }
  return ScanResults;
}

/*************************************************************************
* FUNCTION: ExplainError
*
* ARGS:    hWnd      handle to a display window
*          lpText    ptr to a character string which defines
*                    these error strings.
*          MsgNumber if non zero it represents a message number
                     for direct output.
* RETURNS: none
*
************************************************************************/

void Twain::ExplainError(HWND hWnd, LPSTR lpText, int MsgNumber) {
  DC_STATUS dcStatus;
  DC_UINT16 Status;
  char Details [255];
  int ErrorCode;

  // Put in detail of current operation
  lstrcpy (Details, lpText);

  if (MsgNumber != 0)             // If MsgNumber non zero output specified
    ErrorCode = MsgNumber;        // error message directly without polling
  else {                          // SM.
    // determine details of failure from SM
    if (*lpDSM_Entry) {
      Status = (*lpDSM_Entry)(&AppIdentity,
                            NULL,
                            DG_CONTROL,
                            DAT_STATUS,
                            MSG_GET,
                            (DC_MEMREF)&dcStatus);

      if (Status == DCRC_SUCCESS) {
        switch(dcStatus.ConditionCode) {
          case DCCC_SUCCESS:
            ErrorCode = IDS_SUCCESS;
            break;

          default:
          case DCCC_BUMMER:
            ErrorCode = IDS_BUMMER;
            break;

          case DCCC_LOWMEMORY:
            ErrorCode = IDS_LOWMEMORY;
            break;

          case DCCC_NODS:
            ErrorCode = IDS_NODS;
            break;

          case DCCC_MAXCONNECTIONS:
            ErrorCode = IDS_MAXCONNECTIONS;
            break;

          case DCCC_OPERATIONERROR:
            // Error in the DS or DSM which was already reported to the user.  The
            // App should not be redundant.
            goto NoError;

          case DCCC_BADCAP:
            ErrorCode = IDS_BADCAP;
            break;
     
          case DCCC_BADPROTOCOL:
            ErrorCode = IDS_BADPROTOCOL;
            break;

          case DCCC_BADVALUE:
            ErrorCode = IDS_BADVALUE;
            break;

          case DCCC_SEQERROR:
            ErrorCode = IDS_SEQERROR;
            break;

          case DCCC_BADDEST:
            ErrorCode = IDS_BADDEST;
            break;
        }
      } else {
        ErrorCode = IDS_NODETAIL;
      }
    } else 
      ErrorCode = IDS_NODSM;
  }  
  // Now send out the text in a message box
  LoadString (hDLLInstance, ErrorCode, Details + lstrlen(Details), sizeof(Details));
  MessageBox (hWnd, Details, "TWAIN Error", MB_OK);

  // don't report error
  NoError:
  return;
}

/*************************************************************************
*FUNCTION:RestrictToRGB
*
*ARGS:    None
*
*RETURNS: TRUE if set successful
*
*NOTES:   This function forces the Polariod scanner into doing just
*         color acquisition because the TIFF gray scales files the scanner
*         writes are bogus.
*************************************************************************/

BOOL Twain::RestrictToRGB(void) {
  DC_CAPABILITY Caps;
  DC_ENUMERATION Enum;
  DC_INT16 Status;


  Caps.Cap     = ICAP_PIXELTYPE;  // Set this capability
  Caps.ConType = DCON_ENUMERATION;// With this type of container

  Enum.ItemType = DCTY_UINT16;    // Type of enum data is UINT16
  Enum.NumItems = 1;              // Only a single item in enum
  Enum.CurrentIndex = 0;          // Set indices to 0
  Enum.DefaultIndex = 0;

  // Build up the container
  DC_UINT16 ValidAcqType = DCPT_RGB;
  if (!BuildUpEnumerationType(&Caps, &Enum, &ValidAcqType))
    return FALSE;                  // False if no memory

  // Tell the source
  Status = (*lpDSM_Entry)(&AppIdentity,
                          &SourceIdentity,
                          DG_CONTROL,
                          DAT_CAPABILITY,
                          MSG_SET,
                          (DC_MEMREF)&Caps);

  GlobalFree(Caps.hContainer);     // Free up container memory
  if (Status != DCRC_SUCCESS) {    // If problem occured
    ExplainError(hAppWindow, "RestrictToRGB - ", 0);
    return FALSE;
  }
  return TRUE;                    // Everything OK
}


/*************************************************************************
*FUNCTION:SetXferMechanism
*
*ARGS:    Image transfer mechanism. One of the following:
*         DCSX_NATIVE, DCSX_FILE, DCSX_MEMORY
*
*RETURNS: TRUE if mechanism successfully set
*
*NOTES:   This function should check to see if Mechanism is supported
*         before forcing it.
*************************************************************************/

BOOL Twain::SetXferMechanism(DC_UINT16 Mechanism) {
 DC_CAPABILITY Caps;
 DC_INT16 Status;

 Caps.Cap = ICAP_XFERMECH;        // Set this capability
 if (!BuildUpOneValue(&Caps, DCTY_UINT16, Mechanism))
   return FALSE;                  // False if no memory

 // Tell the source
 Status = (*lpDSM_Entry)(&AppIdentity,
                         &SourceIdentity,
                         DG_CONTROL,
                         DAT_CAPABILITY,
                         MSG_SET,
                         (DC_MEMREF)&Caps);

 GlobalFree(Caps.hContainer);     // Free up container memory
 if (Status != DCRC_SUCCESS) {    // If problem occured
   ExplainError(hAppWindow, "SetXferMechanism - ", 0);
   return FALSE;
 }
 XferMechanism = Mechanism;       // Save type of xfer globally
 return TRUE;                     // Everything OK
}


/*************************************************************************
*FUNCTION:SetUpFileTransfer
*
*ARGS:    FileName is the name to be given to the scanned image file.
*
*RETURNS: TRUE if setup was completed successfully. Also the FileName
*         is given a .TIF file extension.
*
*NOTES:   This function should check to see if TIFF is supported by the
*         TWAIN device before forcing it.
*************************************************************************/

BOOL Twain::SetupFileTransfer(LPSTR FileName) {
  DC_SETUPFILEXFER FileXfer;
  DC_INT16 Status;

  if (!SetXferMechanism(DCSX_FILE))// Ask the device to use file xfer
    return FALSE;                 // If it says no, we're in trouble

  LPSTR PeriodPtr = strchr(FileName,'.');   // Search for period of extension
  if (PeriodPtr != NULL)          // If period found, truncate name there
    *PeriodPtr = '\0';
  lstrcat(FileName, ".tif");      // Add the TIF file extension.

  // Fill in the DC_SETUPFILEXFER structure with the info
  lstrcpy((LPSTR) &FileXfer.FileName, FileName); // Copy the filename to use
  FileXfer.Format = DCFF_TIFF;    // Ask for a TIFF file
  FileXfer.VRefNum = 0;           // This field is DONTCARE

  // Tell the source about the filename and type
  Status = (*lpDSM_Entry)(&AppIdentity,
                          &SourceIdentity,
                          DG_CONTROL,
                          DAT_SETUPFILEXFER,
                          MSG_SET,
                          (DC_MEMREF)&FileXfer);

  if (Status != DCRC_SUCCESS) {   // If error occurred
    ExplainError(hAppWindow, "SetupFileTransfer - ", 0);
    return FALSE;                 // Return FALSE
  }
  return TRUE;
}

/*************************************************************************
*FUNCTION:SetResolution
*
*ARGS:    DPI The requested maximum resolution that the Source should
*             allow the selection of.
*
*
*RETURNS: TRUE if resolution set was successful. FALSE if the device is only
*         capable of one resolution or if device does not negotiate this
*         capability.
*
*NOTES:   All resolutions provided by the device that are <= DPI
*         will be selectable in the Source's UI.
*************************************************************************/
#define MAXRESOLUTIONS 10

BOOL Twain::SetResolution(int DPI) {
  BOOL CanSetResolution = FALSE;  // TRUE if scanner can negotiate
  float UseableResolutions[MAXRESOLUTIONS]; // List of resolutions <= DPI then scanner can use
  DC_FIX32 UseResolutions[MAXRESOLUTIONS];  // Same list in FIX32 type
  float Resolution;
  DC_CAPABILITY Caps;
  DC_INT16 Status;
  DC_RANGE ReturnedRange;
  DC_FIX32 ReturnedFix32;
  int Index;

  // Clear resolution storage area
  memset(UseableResolutions, 0, sizeof(UseableResolutions));

  Caps.Cap = ICAP_XRESOLUTION;    // Get values for this ICAP

  // Ask the source for these values
  Status = (*lpDSM_Entry)(&AppIdentity,
                          &SourceIdentity,
                          DG_CONTROL,
                          DAT_CAPABILITY,
                          MSG_GET,
                          (DC_MEMREF)&Caps);
  if (Status != DCRC_SUCCESS) {   // If error occurred
    GlobalFree(Caps.hContainer);  // Free up container memory
    return FALSE;                 // Return FALSE
  }
  // We have gotten the capability container
  switch(Caps.ConType) {          // Parse the returned container type
    case DCON_ONEVALUE:           // If its one value, were stuck with it
      break;                      // there is nothing we can do.

    case DCON_RANGE:              // A RANGE was returned
      Index = 0;
      ExtractRange(&Caps, &ReturnedRange);  // Extract RANGE from container
      long MinVal = ReturnedRange.MinValue;
      long MaxVal = ReturnedRange.MaxValue;
      long Step   = ReturnedRange.StepSize;
      do {                        // Store all resolutions <= DPI
        UseableResolutions[Index++] = (float) MinVal;
        if (MinVal >= DPI)
          break;
        if (MinVal >= MaxVal)
          break;
        MinVal += Step;
      } while (TRUE);
      CanSetResolution = TRUE;
      break;

    case DCON_ENUMERATION:        // A ENUMERATION was returned
      Index = 0;                  // Index of enumerated value to return
      do {
        if (!ExtractEnumerationValue(&Caps, &ReturnedFix32, Index))
          break;                  // Index exceeded items in enumeration
        Resolution = FIX32ToFloat(ReturnedFix32);
        UseableResolutions[Index++] = Resolution;
        if (Resolution >= (float) DPI)
          break;
      } while (TRUE);
      CanSetResolution = TRUE;
      break;
  }
  GlobalFree(Caps.hContainer);    // Free container memory
  // See if we need to attempt to set the resolution
  if (!CanSetResolution)          // If scanner will not negotiate
    return FALSE;                 // return error

  // Attempt to limit the resolution of the Source device

  WORD NumOfResolutions = 0;
  for (Index = 0; Index < MAXRESOLUTIONS; Index++) {
    if (UseableResolutions[Index] > 0.1) {
      // Count how may resolutions there are <= DPI
      // and convert them from float to FIX32 type.
      NumOfResolutions++;
      UseResolutions[Index] = FloatToFIX32(UseableResolutions[Index]);
    }
  }
  Caps.Cap = ICAP_XRESOLUTION;
  Caps.ConType = DCON_ENUMERATION;
  DC_ENUMERATION Enum;            // Create enumeration data structure
  Enum.ItemType = DCTY_FIX32;     // Type of enum data is FIX32
  Enum.NumItems = NumOfResolutions;    // The number of resolutions
  Enum.CurrentIndex = 0;               // that meet the requirements
  Enum.DefaultIndex = 0;

  // Build up the container
  BuildUpEnumerationType(&Caps, &Enum, &UseResolutions);

  // Ask the source to set this resolution
  Status = (*lpDSM_Entry)(&AppIdentity,
                          &SourceIdentity,
                          DG_CONTROL,
                          DAT_CAPABILITY,
                          MSG_SET,
                          (DC_MEMREF)&Caps);

  if (Status != DCRC_SUCCESS) {   // If error occurred
    ExplainError(hAppWindow, "SetResolution - ", 0);
    return FALSE;
 }
 return TRUE;                     // Everything OK
}


/*************************************************************************
*FUNCTION: OpenDSM
*
*ARGS:    none
*
*RETURNS: current state of the Source Manager
*
*NOTES:     1) Makes sure SM is not already open
*           2) Explicitly loads the DLL for the Source Manager
*           3) Call Source Manager to:
*              - open SM
*              - pass the handle to the app's window to the SM
*              - Set the SM assigned AppIdentity.id field
*************************************************************************/

BOOL Twain::OpenDSM(HWND hWindow) {
  DC_UINT16     Status;
  OFSTRUCT      OpenFiles;
  char          WinDir[WINDIRPATHSIZE];
  DC_STR32      DSMName;


  // Only open SM if currently closed
  if (!DSMOpen) {
    hAppWindow = hWindow;         // Save the App window handle
    // Find the path to the Windows directory and then attempt to open the SM.
    GetWindowsDirectory(WinDir, WINDIRPATHSIZE);
    if (WinDir[lstrlen(WinDir)-1] != '\\')
      lstrcat (WinDir, "\\");

    // Load DSM name from resource and merge into path
    LoadString (hDLLInstance, IDS_DSMNAME, DSMName, sizeof(DC_STR32));
    lstrcat (WinDir, DSMName);

    // Check to see if it exists. If so open library
    if ((OpenFile(WinDir, &OpenFiles, OF_EXIST) == HFILE_ERROR) ||
       ((hDSMDLL = LoadLibrary(DSMName)) == NULL)               ||
	(hDSMDLL <= HINSTANCE_ERROR)                            ||
       ((lpDSM_Entry = (DSMENTRYPROC) GetProcAddress(hDSMDLL, "DSM_ENTRY")) == NULL)) {
      // NOTE: must pass error number here so ExplainError will not attempt
      // to ask DSM what the details of the error were. This is not possible
      // because the DSM could not be opened.
      ExplainError(hWindow, "OpenDSM - ", IDS_NODSM);
      return FALSE;
    }
    /*
    This call performs four important functions:
      - opens/loads the SM
      - passes the handle to the app's window to the SM
      - returns the SM assigned appID.id field
      - be sure to test the return code for SUCCESSful open of SM
    */
    Status = (*lpDSM_Entry)(&AppIdentity,
                          NULL,
                          DG_CONTROL,
                          DAT_PARENT,
                          MSG_OPENDSM,
                          (DC_MEMREF)&hWindow);


    switch (Status) {
      case DCRC_SUCCESS:
        // Open was a success
        DSMOpen = TRUE;
        break;

      case DCRC_FAILURE:
      default:
        // Trouble opening the SM, inform the user
        DSMOpen = FALSE;
        ExplainError(hWindow, "OpenDSM - ", 0);
        break;
    }
  }
  // Let the caller know what happened
  return (DSMOpen);
}



/*************************************************************************
*FUNCTION: CloseDSM
*
*ARGS:    none
*
*RETURNS: TRUE if DSM was closed or was already closed
*
*NOTES:   1) Makes sure SM is already open
*         2) calls Source Manager to request closure from this app.
*************************************************************************/

BOOL Twain::CloseDSM(void) {
  DC_UINT16 Status;

  // Only close something which is already open
  if (DSMOpen) {
    /* This call performs one important function:
       - tells the SM which application, AppIdentity.id, is requesting SM to close
       - be sure to test return code, failure indicates SM did not close !!
    */
    Status = (*lpDSM_Entry)(&AppIdentity,
                          NULL,
                          DG_CONTROL,
                          DAT_PARENT,
                          MSG_CLOSEDSM,
                          &hAppWindow);

    if (Status != DCRC_SUCCESS) {
      ExplainError(hAppWindow, "CloseDSM - ", 0);
      return FALSE;
    } else {
      DSMOpen = FALSE;            // Set flag if close was successful
      // Explicitly free the SM library
      if (hDSMDLL) {
        FreeLibrary (hDSMDLL);
        hDSMDLL=0;
      }
      hAppWindow = 0;             // Clear handle to App window
    }
  }
  return TRUE;
} 


/*************************************************************************
*FUNCTION: OpenDS
*
*ARGS:    none
*
*RETURNS: current state of select Source
*
*NOTES:
*  1) Only attempt to open a source if it is currently closed
*  2) Call Source Manager to:
*     - open the Source indicated by info in SourceIdentity
*     - SM will fill in the unique SourceIdentity.Id field
*************************************************************************/

BOOL Twain::OpenDS(void) {
  DC_UINT16 Status;

  // If a Source is not already open
  if (!DSOpen) {
    // This will open the specified Source and will give it a new ID
    // for this session. If this is called before SelectDS, the default
    // source will be opened.
    SourceIdentity.Id = 0;
    Status = (*lpDSM_Entry)(&AppIdentity,
                          NULL,
                          DG_CONTROL,
                          DAT_IDENTITY,
                          MSG_OPENDS,
                          &SourceIdentity);

    if (Status == DCRC_SUCCESS) {
      // Do not change flag unless we successfully open
      DSOpen = TRUE;
    } else {
      // Trouble opening the Source
      ExplainError (hAppWindow, "OpenDS - ", 0);
    }
  }
  return DSOpen;
}

/*************************************************************************
*FUNCTION: CloseDS
*
*ARGS:    none
*
*RETURNS: none
*
*NOTES:   1) only attempt to close an open Source
*         2) call Source Manager to ask identified Source to close itself
*************************************************************************/

BOOL Twain::CloseDS(void) {
  DC_UINT16 Status;
  
  if (DSOpen) {
    // Close an open Source
    Status = (*lpDSM_Entry)(&AppIdentity,
                          NULL,
                          DG_CONTROL,
                          DAT_IDENTITY,
                          MSG_CLOSEDS,
                          &SourceIdentity);

    if (Status != DCRC_SUCCESS) {
      ExplainError(hAppWindow, "CloseDS - ", 0);
      return FALSE;
    } else {
      DSOpen = FALSE;
    }
  }
  return TRUE;
}

/*************************************************************************
*FUNCTION:EnableDS
*
*ARGS:    none
*
*RETURNS: BOOL for TRUE=open; FALSE=not open/fail
*
*NOTES:   1) only enable an open Source
*         2) call the Source Manager to bring up the Source's User Interface
*************************************************************************/

BOOL Twain::EnableDS (void) {
  BOOL Result=FALSE;
  DC_UINT16 Status;
  DC_USERINTERFACE dcUI;

  // only enable an open Source
  if (DSOpen) {
    // This will display the Source User Interface. The Source should only
    // display a user interface that is compatible with the group defined
    // by AppIdentity.SupportedGroups (in our case DG_IMAGE | DG_CONTROL).
    dcUI.hParent = hAppWindow;
    dcUI.ShowUI  = TRUE;
    Status = (*lpDSM_Entry)(&AppIdentity,
                          &SourceIdentity,
                          DG_CONTROL,
                          DAT_USERINTERFACE,
                          MSG_ENABLEDS,
                          (DC_MEMREF)&dcUI);

    if (Status!=DCRC_SUCCESS)
      ExplainError (hAppWindow, "EnableDS - ", 0);
    else {
      Result = TRUE;
      DSEnabled = TRUE;
    }            
  }
  return Result;
}

/*************************************************************************
*FUNCTION:DisableDS
*
*ARGS:    none
*
*RETURNS: none
*
*NOTES:   1) Only disable an open Source
*         2) Call Source Manager to ask Source to hide it's User Interface
*************************************************************************/

BOOL Twain::DisableDS(void) {
  DC_UINT16 Status;
  DC_USERINTERFACE dcUI;
  BOOL Result = FALSE;

  // only disable open Source's
  if (DSOpen && DSEnabled) {
    // Hide the Source UI
    dcUI.hParent = hAppWindow;
    dcUI.ShowUI  = FALSE;

    Status = (*lpDSM_Entry)(&AppIdentity,
                          &SourceIdentity,
                          DG_CONTROL,
                          DAT_USERINTERFACE,
                          MSG_DISABLEDS,
                          (DC_MEMREF)&dcUI);

    if (Status!=DCRC_SUCCESS)
      ExplainError (hAppWindow, "DisableDS - ", 0);
    else {
      Result = TRUE;
      DSEnabled = FALSE;
    }            
  }
  return Result;
}

/*************************************************************************
*FUNCTION:SelectDS
*
*ARGS:    none
*
*RETURNS: BOOL TRUE if successfull, FALSE if not
*
*NOTES:   1) Call the Source Manager to:
*              - have the SM put up a list of the available Sources
*              - get information about the user selected Source from
*                NewDSIdentity, filled by Source
*************************************************************************/

BOOL Twain::SelectDS(void) {
  DC_UINT16 Status;
  DC_IDENTITY NewDSIdentity;
  BOOL Result = TRUE;
    
  if (DSOpen) {
    ExplainError(hAppWindow, "SelectDS - ", IDS_SOURCEOPEN);
    return FALSE;
  }
  // Get the system default Source
  Status = (*lpDSM_Entry)(&AppIdentity,
	                NULL,
	                DG_CONTROL,
	                DAT_IDENTITY,
	                MSG_GETDEFAULT,
	                (DC_MEMREF)&NewDSIdentity);
	
	
  /* This call performs one important function:
      - should cause SM to put up dialog box of available Source's
      - tells the SM which application, appID.id, is requesting, REQUIRED
      - returns the SM assigned NewDSIdentity.id field, you check if changed
        (needed to talk to a particular Data Source)
      - be sure to test return code, failure indicates SM did not close !!
  */
  Status = (*lpDSM_Entry)(&AppIdentity,
	                NULL,
	                DG_CONTROL,
	                DAT_IDENTITY,
	                MSG_USERSELECT,
	                (DC_MEMREF)&NewDSIdentity);
	
   /* Check if the user changed the Source and react as apporpriate.
      - DCRC_SUCCESS, log in new Source
      - DCRC_CANCEL,  keep the current Source
      - default,      check down the codes in a status message, display result
  */
  switch (Status) {
    case DCRC_SUCCESS:
      SourceIdentity = NewDSIdentity; 
      break;

    case DCRC_CANCEL:
      break;

    default:
    case DCRC_FAILURE:	        
      ExplainError(hAppWindow, "SelectDS - ", 0);
      Result = FALSE;
      break;
  }
  // Let the caller know what happened
  return (Result);
}


/*************************************************************************
*FUNCTION:SelectSource
*
*ARGS:    hWindow is handle of window dialog box should come up in.
*
*RETURNS: BOOL TRUE if successfull, FALSE if not
*
*NOTES:   Opens Source Manager to put up a list of the available Sources
*************************************************************************/

BOOL Twain::SelectSource(HWND hWindow) {

  if (OpenDSM(hWindow)) {         // Open the DSM
    SelectDS();                   // If successful ask it to display sources
    CloseDSM();                   // Close DSM before exit
    return TRUE;                  // Return TRUE for no error
  } else
    return FALSE;                 // FALSE means error occurred.
}


// Backdrop window Window Procedure
long CALLBACK _export Twain::BackDropWndProc(HWND hwnd, UINT message,
                                             WPARAM wParam, LPARAM lParam) {
  static BOOL FirstPaint;
  PAINTSTRUCT Ps;
  RECT Rect;
  static HWND hUIWindow;

  switch (message)   {

    case WM_CREATE:
      // Here is where the image scan is setup
      if (!lpTwain->OpenDSM(hwnd)) return -1L;   // Attempt to open the DSM
      if (!lpTwain->OpenDS()) {   // Attempt to open the source
        lpTwain->CloseDSM();      // If error, close DSM
        return -1L;
      }
      if (!lpTwain->SetResolution(250)) {   // Attempt to set max resolution
        lpTwain->CloseDS();
        lpTwain->CloseDSM();
        return -1L;
      }                           // Request a file xfer
      if (!lpTwain->SetupFileTransfer(lpTwain->lpFileName)) {
        lpTwain->CloseDS();
        lpTwain->CloseDSM();
        return -1L;
      }                           // Restrict scanner to RGB images
      hUIWindow = 0;              // No handle yet for Source UI window
      FirstPaint = TRUE;          // Indicate window not yet painted

      // If we get here all is well
      return 0L;

    case WM_SETFOCUS:             // Background window has been given focus
      if (hUIWindow)
        SetFocus(hUIWindow);      // Return focus to Source's UI
      return 0L;

    case WM_PAINT:                
      BeginPaint(hwnd, &Ps);
      if (FirstPaint) {           // If first time window is painted
        FirstPaint = FALSE;       // Only do this once
        lpTwain->EnableDS();      // Bring up Source's UI
        InvalidateRect(hwnd, NULL, FALSE);  // Force a repaint so code 
      } else {                    // below will be executed
        hUIWindow = GetParent(GetFocus());  // Get handle of Source's window
        GetWindowRect(hUIWindow, &Rect);    // Center it on the screen
        SetWindowPos(hUIWindow, NULL,
          (GetSystemMetrics(SM_CXSCREEN) - (Rect.right  - Rect.left)) / 2,
          (GetSystemMetrics(SM_CYSCREEN) - (Rect.bottom - Rect.top))  / 2,
          0, 0, SWP_NOSIZE | SWP_NOZORDER);
      }
      EndPaint(hwnd, &Ps);
      return 0L;

    case WM_SYSCOMMAND:
      if ((wParam & 0xFFF0) == SC_MOVE)// Don't allow background window to move
        return 0L;
      break;

    case WM_XFERDONE:             // Source says it done xfering image
      if (wParam)                 // If the transfer was successful
        lpTwain->ScanResults = TRUE;// Indicate that an image was scanned
      break;

    case WM_DESTROY:
      PostQuitMessage(0);
      return 0L;
  }
  return DefWindowProc (hwnd, message, wParam, lParam);
}


// Library entry point for initialization
int CALLBACK LibMain (HANDLE hInstance, WORD, WORD wHeapSize,
                      LPSTR)  {

  hDLLInst = hInstance;

  if (wHeapSize > 0)
    UnlockData (0);

  WNDCLASS wndclass;

  // Define the backdrop window class
  wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_BYTEALIGNWINDOW;
  wndclass.lpfnWndProc   = Twain::BackDropWndProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = hInstance;
  wndclass.hIcon         = NULL;
  wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
  wndclass.hbrBackground = GetStockObject(WHITE_BRUSH);
  wndclass.lpszMenuName  = NULL;
  wndclass.lpszClassName = szAppName;
  RegisterClass(&wndclass);

  return 1;
}

// Window's exit procedure for this DLL
int CALLBACK WEP(int nParameter)  {

  if (nParameter == WEP_SYSTEM_EXIT)  {
    // system shutdown in progress
    return(1);
  }
  else if (nParameter == WEP_FREE_DLL)  {
    // Every application has freed this DLL
    // Free up memory for these items
    return(1);
  }  else  {
    // Undefined value -- ignor
    return(1);
  }
}


